# sccsid = "@(#) src/hpcc/common/LDAPUtils.pm, hpc.cloud.comm, hpcc_rbei, rbeis002a 1.4 9/21/11 03:13:03";
# IBM_PROLOG_BEGIN_TAG 
# This is an automatically generated prolog. 
#  
#  
#  
# Licensed Materials - Property of IBM 
#  
# Restricted Materials of IBM 
#  
# (C) COPYRIGHT International Business Machines Corp. 2010,2011 
# All Rights Reserved 
#  
# US Government Users Restricted Rights - Use, duplication or 
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp. 
#  
# IBM_PROLOG_END_TAG 

package LDAPUtils;

# ----------------------------------------------------------------- #
# NOTE:                                                             #
#                                                                   #
# All methods (except of new()) will return result in this way:     #
#   ($rc, $param)                                                   #
# $rc : indicates the return code; 0 or non-0                       #
#       0     : succ, the $param will take proper result;           #
#       non-0 : fail, the $param will be error msg;                 #
#                                                                   #
# ----------------------------------------------------------------- #


use strict;
use warnings;
use Net::LDAP;
use Digest::SHA qw(sha1_base64);
use Utils::Logger;

# -------------------------------------------------------------------
# The information should be read out from configuration
# -------------------------------------------------------------------
$LDAPUtils::ManagerDN = 'cn=Manager,dc=thcloud.com';	# manager DN
$LDAPUtils::UserDN = 'uid=%s,ou=people,dc=thcloud.com';	# user DN
$LDAPUtils::PeopeDn = 'ou=people,dc=thcloud.com';	# peope node DN
$LDAPUtils::UidFilter = '(uid=%s)';			# user filter
$LDAPUtils::PeopleObjectClass = ['account', 'posixAccount', 'shadowAccount', 'top'];
							# default peope class
$LDAPUtils::UserAttrMask = ['uid', 'uidNumber', 'gidNumber', 'homeDirectory', 'loginShell', 'host', 'objectclass', 'userPassword'];
							# only those user attributes will be retrived;
$LDAPUtils::InternelDebug = 0;

sub new {
	my $class = shift;
	my $attr = shift;
	my $self = {
		_conn_host => $attr->{'host'},
		_conn_port => $attr->{'port'},
		_conn_uid => $attr->{'dn'},
		_conn_pw => $attr->{'pw'},
		_hdl => 0,
	};

	bless $self, $class;
	return $self;
}

# get user DN;
sub userDN {
	my ($uid) = @_;
	if (index($uid, ',') == -1) {
		return sprintf($LDAPUtils::UserDN, $uid);
	} else {
		return $uid;
	}
}

# get manager DN;
sub managerDN {
	return $LDAPUtils::ManagerDN;
}

# Hash of input password
sub hashPasswordBySHA1 {
	my ($input) = @_;
	my $result = '{SHA}' . sha1_base64($input) . '=';
	$LDAPUtils::InternelDebug && printf("[LDAPUtils::hashPasswordBySHA1::DEBUG] HASH CODE %s for %s\n", $result, $input);
	$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'hashPasswordBySHA1', "HASH CODE $result for $input");

	return $result;
}

# bind as input
sub bind {
	my ($self) = @_;
	if (defined($self->{_conn_uid})) {
		my ($uid, $pw) = ($self->{_conn_uid}, $self->{_conn_pw});
		my $dn = &userDN($uid);
		return $self->bindAs($dn, $pw);
	}

	return $self->bindAs();
}

# bind
sub bindAs {
	my ($self, $dn, $pw) = @_;
	my $rc;
	my $errmsg;
	my $hdl = $self->{_hdl};

	if (defined($dn)== 0) {
		# bind anonymous
		$LDAPUtils::InternelDebug && printf("[LDAPUtils::bind::DEBUG] anonymous bind.(%s:%d)\n", __FILE__, __LINE__);
		my $tmp = sprintf("anonymous bind.(%s:%d)", __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'bindAs', "$tmp");
		$rc = $hdl->bind;
	} else {
		$LDAPUtils::InternelDebug && printf("[LDAPUtils::bind::DEBUG] bind as %s.(%s:%d)\n", $dn, __FILE__, __LINE__);
		my $tmp = sprintf(" bind as %s.(%s:%d)", $dn, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'bindAs', "$tmp");
		$rc = $hdl->bind(&userDN($dn), password => $pw);
	}

	if ($rc->code) {
		$errmsg = $rc->error;
		$LDAPUtils::InternelDebug && printf("[LDAPUtils::bind::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		my $tmp = sprintf("Error: %s (%s:%d)", $errmsg, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'bindAs', "$tmp");
		return (1, $errmsg);
	}

	return (0);
}

# unbind, break the connection;
sub unbind {
	my ($self) = @_;
	my $hdl = $self->{_hdl};
	return $hdl->unbind();
}

# connect to LDAP server
sub connect {
	my ($self) = @_;
	my $host = $self->{_conn_host};
	my $port = $self->{_conn_port};
	my $hdl;
	my $errmsg = '';

	if (defined($port)) {
		$hdl = Net::LDAP->new($host, port=>$port);
		unless ($hdl) {
			$errmsg = $@;
			$LDAPUtils::InternelDebug && printf("[LDAPUtils::connect::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
			my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
			$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'connect', "$tmp");
			return (1, $errmsg);
		}
	} else {
		$hdl = Net::LDAP->new($host);
		unless ($hdl) {
			$errmsg = $@;
			$LDAPUtils::InternelDebug && printf("[LDAPUtils::connect::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
			my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
			$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'connect', "$tmp");
			return (1, $errmsg);
		}
	}

	$self->{_hdl} = $hdl;
	return (0);
}

# Search users under the peope node DN with the input filter;
sub searchUserInfo {

	my ($self, $filter) = @_;
	my $hdl = $self->{_hdl};
	my $errmsg = '';

	my %users = ();

	my $baseDn = $LDAPUtils::PeopeDn;
	my $attrs = $LDAPUtils::UserAttrMask;

	my $mesg = $hdl->search(base => $baseDn, filter => $filter, attrs => $attrs);
	if ($mesg->code) {
		$errmsg = $mesg->error;
		$LDAPUtils::InternelDebug && printf("[LDAPUtils::searchUserInfo::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'searchUserInfo', "$tmp");
		return (1, $errmsg);
	}

	foreach my $entry ($mesg->entries) {
		my %userInfo = ();
		my $userInfoHdl = $entry->{'asn'}->{'attributes'};

		foreach my $fieldHdl (@$userInfoHdl) {
			my $key = $fieldHdl->{'type'};
			my $val = $fieldHdl->{'vals'};

			if ($key eq 'objectClass') {
				# store the people class;
				$LDAPUtils::PeopleObjectClass = $val;
			} else {
				$userInfo{$key} = $val;
			}
		}

		my $userNameHdl = $userInfo{'uid'};
		my $userName = $userNameHdl->[0];
		$users{$userName} = \%userInfo;
	}

	return (0, \%users);
}

# Search one user info;
sub getUserInfo {
	my ($self, $uid) = @_;
	my $filter = sprintf($LDAPUtils::UidFilter, $uid);
	return $self->searchUserInfo($filter);
}

# Search all users info;
sub getAllUserInfo {
	my ($self) = @_;
	my $filter = sprintf($LDAPUtils::UidFilter, '*');
	return $self->searchUserInfo($filter);
}

# Search next user id;
sub getNextUserId {
	my ($self) = @_;
	my $max = 0;

	my ($rc, $users) = $self->getAllUserInfo();

	if ($rc != 0) {
		return (-1, $users);
	}

	foreach my $u (keys(%$users)) {
		my $info = $users->{$u};
		my $uid = $info->{'uidNumber'}->[0];
		$max=$uid if ($uid > $max);
	}

	$max++;

	return (0, $max);
}

# Add one new user
sub addUser {
	my ($self, $uid, $atts) = @_;
	my $hdl = $self->{_hdl};
	my $rc;
	my $errmsg = '';
	my $dn = sprintf('uid=%s,ou=people,dc=thcloud.com', $uid);
	my $uidNumber = $self->getNextUserId();
	my $new_atts = ['cn'=>[$uid], 'uid'=>[$uid], 'uidNumber'=>[$uidNumber], 'objectclass'=>$LDAPUtils::PeopleObjectClass];

	foreach (keys(%$atts)) {
		push(@$new_atts, $_);
		push(@$new_atts, $atts->{$_});
	}

	$rc = $hdl->add($dn, attr => $new_atts);
	if ($rc->code) {
		$errmsg = $rc->error;
		$LDAPUtils::InternelDebug &&  printf("[LDAPUtils::addUser::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'addUser', "$tmp");
		return (1, $errmsg);
	}

	return (0);
}

sub delUser {
	my ($self, $uid) = @_;
	my $hdl = $self->{_hdl};
	my $rc;
	my $errmsg = '';
	my $dn = sprintf('uid=%s,ou=people,dc=thcloud.com', $uid);
	
	$rc = $hdl->delete($dn);
	if ($rc->code) {
		$errmsg = $rc->error;
		$LDAPUtils::InternelDebug &&  printf("[LDAPUtils::delUser::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'delUser', "$tmp");
		return (1, $errmsg);
	}

	return (0);
}

# Modify one user
sub modUser {
	my ($self, $uid, $attr) = @_;
	my $hdl = $self->{_hdl};
	my $rc;
	my $errmsg = '';
	my $dn = sprintf('uid=%s,ou=people,dc=thcloud.com', $uid);

	$rc = $hdl->modify($dn, replace => $attr);
	if ($rc->code) {
		$errmsg = $rc->error;
		$LDAPUtils::InternelDebug && printf("[LDAPUtils::modUser::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'modUser', "$tmp");
		return (1, $errmsg);
	}

	return (0);
}

# Modify one user's password
sub modUserPassword {
	my ($self, $uid, $newpw) = @_;
	my $hdl = $self->{_hdl};
	my $rc;
	my $errmsg = '';
	my $dn = sprintf('uid=%s,ou=people,dc=thcloud.com', $uid);

	my $encoded = &hashPasswordBySHA1($newpw);

	$rc = $hdl->modify($dn, replace => {'userPassword'=>[$encoded]});
	if ($rc->code) {
		$errmsg = $rc->error;
		$LDAPUtils::InternelDebug && printf("[LDAPUtils::modUser::DEBUG] Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		my $tmp = sprintf("Error: %s (%s:%d)\n", $errmsg, __FILE__, __LINE__);
		$LDAPUtils::InternelDebug && Logger->get()->debug('LDAPUtils', 'modUserPassword', "$tmp");
		return (1, $errmsg);
	}

	return (0);
}

# Modify user's shell
sub modUserShell {
	my ($self, $uid, $newShell) = @_;
	my $hdl = $self->{_hdl};
	my $result;
	my $errmsg = '';
	printf ("In LDAPUtils::modUserShell: user=${uid}, shell=${newShell}\n");
	my $dn = sprintf('uid=kmarthi,ou=people,dc=thcloud.com', $uid);

	$result = $hdl->modify($dn, replace => {'loginShell'=>['/bin/bash']});
	$result->code && warn "failed to replace loginShell entry: ", $result->error ;

	return (0);
}

1;
